@@ -139,8 +139,8 @@ def likelihood_scan(**kwargs):
               help='evaluate asymptotic CLs limit for negative sigma bands')
 @click.option('--binned/--unbinned', 'binned_likelihood', default=True, help='Binned likelihood')
 @click.option('--save_summary/--skip_summary', default=True, help='Save summary information')
-@click.option('-f', '--fix', default="", help='Parameters to fix')
-@click.option('-r', '--profile', default="", help='Parameters to profile')
+@click.option('-f', '--fix', 'fix_param', default="", help='Parameters to fix')
+@click.option('-r', '--profile', 'profile_param', default="", help='Parameters to profile')
 @click.option('-w', '--workspace', 'ws_name', default=None, help='Name of workspace. Auto-detect by default.')
 @click.option('-m', '--model_config', 'mc_name', default=None, help='Name of model config. Auto-detect by default.')
 @click.option('-s', '--snapshot', 'snapshot_name', default=None, help='Name of initial snapshot')
@@ -162,10 +162,10 @@ def cls_limit(**kwargs):
     """
     Tool for evaluating Asymptotic CLs limit
     """
-    from quickstats.components import AsymptoticsCLs
+    from quickstats.components import AsymptoticCLs
     outname = kwargs.pop('outname')
     save_summary = kwargs.pop('save_summary')
-    asymptotic_cls = AsymptoticsCLs(**kwargs)
+    asymptotic_cls = AsymptoticCLs(**kwargs)
     asymptotic_cls.evaluate_limits()
     asymptotic_cls.save(outname, summary=save_summary)
 
