@@ -12,7 +12,7 @@ from quickstats.utils import common_utils
 import ROOT
 
 from pdb import set_trace
-class AsymptoticsCLs(AnalysisObject):
+class AsymptoticCLs(AnalysisObject):
     
     _SIGMA_LABEL_ = {
             2: '+2sigma',
