@@ -3,7 +3,7 @@
 # Author: Alkaid Cheng
 # Email: chi.lung.cheng@cern.ch
 ##################################################################################################
-from typing import Union, List, Dict
+from typing import Union, Optional, List, Dict
 import math
 
 import numpy as np
@@ -14,7 +14,7 @@ class ExtendedMinimizer(object):
     _NLL_COMMANDS_ = ['NumCPU', 'Constrain', 'CloneData', 'GlobalObservables', 
                       'GlobalObservablesTag', 'OffsetLikelihood', 'Offset']
 
-    _DEFAULT_MINIMIZER_CONFIG_ = {
+    _DEFAULT_MINIMIZER_OPTION_ = {
         'minimizer_type': 'Minuit2',
         'minimizer_algo': 'Migrad',        
         'opt_const': 1,
@@ -38,13 +38,22 @@ class ExtendedMinimizer(object):
         'default_strategy': 0,
         'print_level': 1   
     }
-
+    
+    _DEFAULT_NLL_OPTION_ = {
+        'num_cpu': 1,
+        'offset': True
+    }
+    
     @property
     def config(self):
         return self._config
     
-    def __init__(self, minimizer_name:str, pdf:"ROOT.RooAbsPdf", data:"ROOT.RooAbsData"):
-        
+    def __init__(self, minimizer_name:str, pdf:"ROOT.RooAbsPdf", data:"ROOT.RooAbsData",
+                 minimizer_options:Optional[Dict]=None):
+        """
+            Minimizr Options:
+                precision: precision for the find sigma algorithm
+        """
         self.name = minimizer_name
         self.pdf = pdf
         self.data = data
@@ -52,7 +61,9 @@ class ExtendedMinimizer(object):
         self.nll = None
         self.fit_result = None
         
-        self._config = self._DEFAULT_MINIMIZER_CONFIG_   
+        self._config = self._DEFAULT_MINIMIZER_OPTION_ 
+        if minimizer_options is not None:
+            self._config.update(minimizer_options)
         self.minos_set = ROOT.RooArgSet()
         self.cond_set  = ROOT.RooArgSet()
         self.scan_set  = ROOT.RooArgSet()         
@@ -126,7 +137,7 @@ class ExtendedMinimizer(object):
             raise RuntimeError('NLL not initialized')
 
         if (self.minimizer is None) or (not self.config['reuse_minimizer']):
-            self.configure_default_minimizer_options()            
+            self.configure_default_minimizer_options()
             self.minimizer = ROOT.RooMinimizer(self.nll)
             self.minimizer.setPrintLevel(self.config['print_level'])
             self.minimizer.optimizeConst(self.config['opt_const'])
@@ -137,6 +148,7 @@ class ExtendedMinimizer(object):
             self.minimizer.setProfile(self.config['timer'])
             self.minimizer.setStrategy(self.config['default_strategy'])
             self.minimizer.setEps(self.config['eps'])
+            
             if self.config['max_calls'] != -1:
                 self.minimizer.setMaxFunctionCalls(self.config['max_calls'])
             if self.config['max_iters'] != -1:
@@ -148,6 +160,18 @@ class ExtendedMinimizer(object):
             nll_command_list.Add(command)
         return nll_command_list
     
+    def configure_nll(self, constrain=None, global_observables=None, **kwargs):
+        commands = []
+        if 'num_cpu' in kwargs:
+            commands.append(ROOT.RooFit.NumCPU(kwargs['num_cpu'], 3))
+        if 'offset' in kwargs:
+            commands.append(ROOT.RooFit.Offset(kwargs['offset']))
+        if constrain is not None:
+            commands.append(ROOT.RooFit.Constrain(constrain))
+        if global_observables is not None:
+            commands.append(ROOT.RooFit.GlobalObservables(global_observables))            
+        self.nll_command_list = self.parse_nll_commands(commands)
+    
     def configure(self, **kwargs):
         minos_set = kwargs.pop('minos_set', None)
         cond_set  = kwargs.pop('cond_set' , None)
@@ -172,10 +196,13 @@ class ExtendedMinimizer(object):
         self.fit_options = kwargs
         self.scan_options = {k:v for k,v in kwargs.items() if k not in self._SCAN_EXCLUDE_CONFIG_}
 
-    def minimize(self, **kwargs):
+    def minimize(self, nll=None, **kwargs):
 
         self.configure(**kwargs)
-        self.create_nll()
+        if nll is None:
+            self.create_nll()
+        else:
+            self.nll = nll
         
         if self.cond_set.size() > 0:
             attached_set = self.nll.getVariables()
@@ -270,6 +297,7 @@ class ExtendedMinimizer(object):
         self.eigen_vectors.Print()
     
     def robust_minimize(self):
+
         strategy = self.config['default_strategy']
         retry = self.config['retry']
         minimizer_type = self.config['minimizer_type']
