@@ -1,12 +1,12 @@
 ##################################################################################################
-# Based on https://gitlab.cern.ch/atlas-physics/stat/tools/StatisticsTools by Stefan Gadatsch
+# Based on https://gitlab.cern.ch/atlas-physics/stat/tools/StatisticsTools
 # Author: Alkaid Cheng
 # Email: chi.lung.cheng@cern.ch
 ##################################################################################################
 import os
 import math
 import fnmatch
-from typing import List, Optional
+from typing import List, Optional, Union, Dict
 
 import ROOT
 
@@ -15,14 +15,29 @@ from quickstats.components.numerics import is_integer
 from quickstats.utils.root_utils import load_macro
 
 class ExtendedModel(object):
-    def __init__(self, fname:str, ws_name:Optional[str]=None, model_config_name:Optional[str]=None,
-                 data_name:str="combData", snapshot_name:Optional[str]="nominalNuis",
+    
+    _DEFAULT_NAMES_ = {
+        'conditional_globs': 'conditionalGlobs_{mu}',
+        'conditional_nuis': 'conditionalNuis_{mu}',
+        'nominal_globs': 'nominalGlobs',
+        'nominal_nuis': 'nominalNuis',
+        'temp_globs': 'tmpGlobs',
+        'temp_nuis': 'tmpNuis',
+        'weight': 'weightVar',
+        'dataset_args': 'obsAndWeight',
+        'asimov': 'asimovData_{mu}',
+        'channel_asimov': 'combAsimovData{index}',
+        'nll_snapshot': '{nll_name}_{mu}'
+    }
+    
+    def __init__(self, fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None,
+                 data_name:str="combData", snapshot_name:Optional[str]=None,
                  binned_likelihood:bool=True, tag_as_measurement:str="pdf_",
                  fix_cache:bool=True, fix_multi:bool=True, interpolation_code:int=-1,
                  load_extension:bool=True):
         self.fname = fname
         self.ws_name = ws_name
-        self.mc_name = model_config_name
+        self.mc_name = mc_name
         self.data_name = data_name
         self.snapshot_name = snapshot_name
         self.binned_likelihood = binned_likelihood
@@ -71,6 +86,11 @@ class ExtendedModel(object):
                     print('INFO: Loaded extension module "RooTwoSidedCBShape"')
         except Exception as e:
             print(e)
+            
+    @staticmethod
+    def load_RooDataSet_dict():
+        load_macro('RooDataSetObjects')
+        
     @staticmethod
     def modify_interp_codes(ws, interp_code, classes=None):
         if classes is None:
@@ -116,7 +136,7 @@ class ExtendedModel(object):
     def initialize(self):
         if not os.path.exists(self.fname):
             raise FileNotFoundError('workspace file {} does not exist'.format(self.fname))
-        print('Opening file: {}'.format(self.fname))
+        print('INFO: Opening file "{}"'.format(self.fname))
         file = ROOT.TFile(self.fname) 
         if (not file):
             raise RuntimeError("Something went wrong while loading the root file: {}".format(self.fname))
@@ -399,6 +419,69 @@ class ExtendedModel(object):
                     components.add(pdf)
                     continue
                 find_unique_prod_components(pdf, components, recursion_count+1)
+                
+    @staticmethod 
+    def _unfold_constraints(source:ROOT.RooArgSet, obs:ROOT.RooArgSet, nuis:ROOT.RooArgSet, 
+                           result:ROOT.RooArgSet=None, recursion:int=0, threshold:int=50):
+        if recursion > threshold:
+            raise RuntimeError("failed to find unfold constraints: recusion limit exceeded")
+        if result is None:
+            result = ROOT.RooArgSet()
+        for pdf in source:
+            class_name = pdf.ClassName()
+            if class_name not in ["RooGaussian", "RooLognormal", "RooGamma", "RooPoisson", "RooBifurGauss"]:
+                constraint_set = pdf.getAllConstraints(obs.Clone(), nuis.Clone(), ROOT.kFALSE)
+                ExtendedModel._unfold_constraints(constraint_set, obs, nuis, result=result, recursion=recursion+1)
+            else:
+                result.add(pdf)
+        return result
+    
+    def unfold_constraints(self, threshold:int=50):
+        source = self.get_all_constraints()
+        return self._unfold_constraints(source, self.observables, self.nuisance_parameters, threshold=threshold)
+    
+    @staticmethod
+    def _pair_nuis_and_glob_obs(constraint_set:ROOT.RooArgSet, globs:ROOT.RooArgSet, nuis:ROOT.RooArgSet):
+        nuis_list = []
+        glob_list = []
+        for pdf in constraint_set:
+            target_np   = None
+            target_glob = None
+            components = pdf.getComponents()
+            components.remove(pdf)
+            if components.getSize():
+                for c1 in components:
+                    for c2 in components:
+                        if c1 == c2:
+                            continue
+                        if c2.dependsOn(c1):
+                            components.remove(c1)
+                if (components.getSize() > 1):
+                    raise RuntimeError("failed to isolate proper nuisance parameter")
+                elif (components.getSize() == 1):
+                    target_np = components.first()
+            else:
+                for np in nuis:
+                    if pdf.dependsOn(np):
+                        target_np = np
+                        break
+            if not target_np:
+                print('WARNING: Could not find nuisance parameter for the constrain: {}'.format(pdf.GetName()))
+                continue
+            for glob in globs:
+                if pdf.dependsOn(glob):
+                    target_glob = glob
+                    break
+            if not target_glob:
+                print('WARNING: Could not find global observable for the constrain: {}'.format(pdf.GetName()))
+                continue
+            nuis_list.append(target_np)
+            glob_list.append(target_glob)
+        return nuis_list, glob_list
+    
+    def pair_nuis_and_glob_obs(self):
+        constraint_set = self.unfold_constraints()
+        return self._pair_nuis_and_glob_obs(constraint_set, self.global_observables, self.nuisance_parameters)
     
     def get_all_constraints(self):
         all_constraints = ROOT.RooArgSet()
@@ -410,8 +493,8 @@ class ExtendedModel(object):
             all_constraints.add(constr)
         else:
             # load information needed to determine attributes from ModelConfig 
-            obs = deepcopy(self.observables)
-            nuis = deepcopy(self.nuisance_parameters)
+            obs = self.observables.Clone()
+            nuis = self.nuisance_parameters.Clone()
             all_constraints = self.pdf.getAllConstraints(obs, nuis, ROOT.kFALSE)
             
         # take care of the case where we have a product of constraint terms
@@ -485,6 +568,194 @@ class ExtendedModel(object):
                 nuis.setError(prefit_variation)
                 nuis.removeRange()    
         return None
+    
+    def get_poi(self, poi_name:str, strict:bool=False):
+        poi = self.workspace.var(poi_name)
+        if not poi:
+            raise RuntimeError('workspace does not contain the variable "{}"'.format(poi_name))
+        if strict and (poi not in list(self.pois)):
+            raise RuntimeError('workspace variable "{}" is not part of the POIs'.format(poi_name))
+        return poi
+    
+    def _load_obs_and_weight(self, obs_and_weight:Optional[Union["ROOT.RooArgSet",str]]=None, 
+                             weight_var:Optional[Union["ROOT.RooRealVar",str]]=None):
+        # get the weight variable
+        if weight_var is None:
+            weight_name = self._DEFAULT_NAMES_['weight']
+            weight_var = self.workspace.var(weight_name)
+            if not weight_var:
+                weight_var = ROOT.RooRealVar(weight_name, weight_name, 1)
+                getattr(self.workspace, "import")(weight_var)
+        elif isinstance(weight_var, str):
+            weight_var = self.workspace.var(weight_var)
+            if not weight_var:
+                raise RuntimeError('weight variable "{}" not found in workspace'.format(weight_var))
+        elif not isinstance(weight_var, ROOT.RooRealVar):
+            raise ValueError('weight variable must be of RooRealVar type')
+                
+        # get the obs_and_weight arg set
+        if obs_and_weight is None:
+            default_name = self._DEFAULT_NAMES_['dataset_args']
+            obs_and_weight = self.workspace.set(default_name)
+            if not obs_and_weight:
+                obs_and_weight = ROOT.RooArgSet()
+                obs_and_weight.add(self.observables)            
+                obs_and_weight.add(weight_var)
+                self.workspace.defineSet(default_name, obs_and_weight)
+        elif isinstance(obs_and_weight, str):
+            obs_and_weight = self.workspace.set(obs_and_weight)
+            if not obs_and_weight:
+                raise RuntimeError('named set "{}" not found in workspace'.format(obs_and_weight))
+        elif not isinstance(obs_and_weight, ROOT.RooArgSet):
+            raise ValueError('the argument "obs_and_weight" must be of RooArgSet type')
+        return obs_and_weight, weight_var
+    
+    @staticmethod
+    def get_dataset_map(dataset_dict:Dict):
+        ExtendedModel.load_RooDataSet_dict()
+        dsmap = ROOT.std.map('string, RooDataSet*')()
+        dsmap.keepalive = list()
+        for c, d in dataset_dict.items():
+            dsmap.keepalive.append(d)
+            dsmap.insert(dsmap.begin(), ROOT.std.pair("const string,RooDataSet*")(c, d))
+        return dsmap
+    
+    def generate_asimov_from_pdf(self, name:str="asimovData", pdf:Optional["ROOT.RooAbsPdf"]=None,
+                                 obs_and_weight:Optional[Union["ROOT.RooArgSet",str]]=None, 
+                                 weight_var:Optional[Union["ROOT.RooRealVar",str]]=None,
+                                 extra_args=None):
+        
+        pdf = pdf if pdf is not None else self.pdf
+        if isinstance(pdf, ROOT.RooSimultaneous):
+            raise ValueError("this method should not be called from a simultaneous pdf")
+        
+        obs_and_weight, weight_var = self._load_obs_and_weight(obs_and_weight, weight_var)
+        
+        # get the combined arg set for the asimov dataset
+        arg_set = ROOT.RooArgSet()
+        arg_set.add(obs_and_weight)
+        if extra_args is not None:
+            if isinstance(extra_args, list):
+                for arg in extra_args:
+                    arg_set.add(arg)
+            else:
+                arg_set.add(extra_args)
+                
+        asimov_data = ROOT.RooDataSet(name, name, arg_set, ROOT.RooFit.WeightVar(weight_var))
+        
+        # generate observables defined by the pdf associated with this state
+        obs = pdf.getObservables(self.observables)        
+        target_obs = obs.first()
+        expected_events = pdf.expectedEvents(obs)
+        for i in range(target_obs.numBins()):
+            target_obs.setBin(i)
+            norm = pdf.getVal(obs)*target_obs.getBinWidth(i)
+            n_events = norm*expected_events
+            if n_events <= 0:
+                print("WARNING: Detected bin with zero expected events ({})! Please check"
+                      "your inputs. Obs = {}, bin = {}".format(n_events, target_obs.GetName(), i))
+            elif (n_events > 0) and (n_events < 1e18):
+                asimov_data.add(self.observables, n_events)
+
+        if (asimov_data.sumEntries() != asimov_data.sumEntries()):
+            raise RuntimeError("asimov data sum entries is nan")
+        return asimov_data
+
+    def generate_asimov(self, poi_name:str, poi_val:float, poi_profile:Optional[float]=None,
+                        do_conditional:bool=False, do_import:bool=True, dataset:Optional["ROOT.RooDataSet"]=None,
+                        minimizer_options:Optional[Dict]=None, nll_options:Optional[Dict]=None,
+                        object_names:Optional[Dict]=None):
+        # define simplified ws variable names
+        ws = self.workspace
+        globs = self.global_observables
+        nuis  = self.nuisance_parameters
+        
+        # define names used for various objects
+        names = self._DEFAULT_NAMES_
+        if object_names is not None:
+            names.update(object_names)
+        tmp_glob_name = names['temp_globs']
+        tmp_nuis_name = names['temp_nuis'] 
+        nom_glob_name = names['nominal_globs']
+        nom_nuis_name = names['nominal_nuis']
+        con_glob_name = names['conditional_globs']
+        con_nuis_name = names['conditional_nuis']
+        asimov_data_name = names['asimov'].format(mu=poi_val)
+        channel_asimov_data_name = names['channel_asimov']
+        
+        poi = self.get_poi(poi_name)
+        if not nuis:
+            nuis = ROOT.RooArgSet()
+        nuis_list, glob_list = self.pair_nuis_and_glob_obs()
+
+        ws.saveSnapshot(tmp_glob_name, globs)
+        ws.saveSnapshot(tmp_nuis_name, nuis)
+        if not ws.getSnapshot(nom_glob_name):
+            print('INFO: Saving snapshot "{}"'.format(nom_glob_name))
+            ws.saveSnapshot("nominalGlobs", globs)
+        if not ws.getSnapshot(nom_nuis_name):
+            print('INFO: Saving snapshot "{}"'.format(nom_nuis_name))
+            ws.saveSnapshot("nominalNuis", nuis)
+
+        if do_conditional and (poi_profile is not None):
+            from quickstats.components import ExtendedMinimizer
+            if dataset is None:
+                dataset = self.data
+            minimizer = ExtendedMinimizer("Minimizer", self.pdf, dataset)
+            if minimizer_options is None:
+                minimizer_options = minimizer._DEFAULT_MINIMIZER_OPTION_
+            if nll_options is None:
+                nll_options = minimizer._DEFAULT_NLL_OPTION_
+            minimizer.configure(**minimizer_options)
+            minimizer.configure_nll(self.nuisance_parameters, self.global_observables,
+                                    **nll_options)
+            poi.setVal(poi_profile)
+            poi.setConstant(1)
+            minimizer.minimize()
+        poi.setVal(poi_val)
+        poi.setConstant(0)
+
+        for np, glob in zip(nuis_list, glob_list):        
+            glob.setVal(np.getVal())
+
+        if do_conditional and (poi_profile is not None):
+            ws.saveSnapshot(con_glob_name.format(mu=poi_profile), globs)
+            ws.saveSnapshot(con_nuis_name.format(mu=poi_profile), nuis)
+        else:
+            ws.saveSnapshot(con_glob_name.format(mu=poi_val), globs)
+            ws.saveSnapshot(con_nuis_name.format(mu=poi_val), nuis)
+            
+        if not do_conditional:
+            ws.loadSnapshot(nom_glob_name)
+            ws.loadSnapshot(nom_nuis_name)    
+        poi.setVal(poi_val)
+        
+        sim_pdf = self.pdf
+        if not isinstance(sim_pdf, ROOT.RooSimultaneous):
+            asimov_data = self.generate_asimov_from_pdf(asimov_data_name, sim_pdf)
+        else:
+            asimov_data_map = {}
+            channel_cat = sim_pdf.indexCat()
+            n_cat = len(channel_cat)
+            for i in range(n_cat):
+                channel_cat.setIndex(i)
+                label = channel_cat.getLabel()
+                pdf_cat = sim_pdf.getPdf(label)
+                name = channel_asimov_data_name.format(index=i)
+                asimov_data_map[label] = self.generate_asimov_from_pdf(name, pdf_cat, extra_args=channel_cat)
+
+            obs_and_weight, weight_var = self._load_obs_and_weight()
+            dataset_map = ExtendedModel.get_dataset_map(asimov_data_map)
+            asimov_data = ROOT.RooDataSet(asimov_data_name, asimov_data_name, 
+                                          ROOT.RooArgSet(obs_and_weight, channel_cat),
+                                          ROOT.RooFit.Index(channel_cat),
+                                          ROOT.RooFit.Import(dataset_map),
+                                          ROOT.RooFit.WeightVar(weight_var))
+        if do_import:
+            getattr(ws, "import")(asimov_data)
+        ws.loadSnapshot(nom_glob_name)
+        return asimov_data
+    
     @staticmethod
     def to_dataframe(args):
         import pandas as pd
