@@ -36,7 +36,7 @@ def evaluate_nll(input_file:str, poi_val:float, poi:str="", vmin:float=10., vmax
                 print('INFO: Cached NLL for POI "{}" at {:.2f}: {}'.format(poi_name, poi_val, nll))
             return nll
     start = time.time()
-    model = ExtendedModel(fname=input_file, ws_name=workspace, model_config_name=model_config,
+    model = ExtendedModel(fname=input_file, ws_name=workspace, mc_name=model_config,
                           data_name=data, binned_likelihood=binned, snapshot_name=snapshot, 
                           fix_cache=fix_cache, fix_multi=fix_multi)
     if fix:
